////////////////////////////////////////////////////////////////////////////////
// File:       PegtopCompilers.pas
// Classes:    TPegtopCompilerData, TPegtopCompilerCode
// Version:    1.01
// Date:       04 Sep 2003 created 1.00
//             10 Feb 2005 modified 1.01 (memory leak removed)
// Author:     Jens Gruschel (GRU)
// Copyright:  (c) 2003 Jens Gruschel
// Website:    http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////
// Description:
// TPegtopCompilerData is a container of dynamic size for data of different
// types.
// TPegtopCompilerCode is a container of dynamic size for machine code. It can
// be filled with machine code and is base class of TPegtopAsmCode (see
// PegtopAssembler.pas) and has functions for jumps to labels.
////////////////////////////////////////////////////////////////////////////////
// License:
// Please read the license.txt file that comes with this software
// and visit http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////
// Documentation / Installation:
// Please read the readme.txt file that comes with this software
// and visit http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////

unit PegtopCompilers;

interface

uses
  Classes, SysUtils, Contnrs;

type
  EPegtopCompiler = class(Exception);

  TPegtopVirtualAddr = Integer; // address relative to some start address

  TPegtopCompilerDataType = (cdtByte, cdtDouble, cdtInt64, cdtLongint, cdtSmallint, cdtWord);

  TPegtopCompilerData = class
  private
    FData: packed array of Byte;
    FSize: Integer;
    FAlign: array[TPegtopCompilerDataType] of Integer;
    function GetPtr(A: TPegtopVirtualAddr): Pointer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear; virtual;
    function AddByte(V: Byte): TPegtopVirtualAddr;
    function AddDouble(V: Double): TPegtopVirtualAddr;
    function AddInt64(V: Int64): TPegtopVirtualAddr;
    function AddLongint(V: Longint): TPegtopVirtualAddr;
    function AddSmallint(V: Smallint): TPegtopVirtualAddr;
    function AddWord(V: Word): TPegtopVirtualAddr;
    function IncSize(Bytes, Align: Integer): TPegtopVirtualAddr;
    procedure SetByte(A: TPegtopVirtualAddr; V: Byte);
    property Size: Integer read FSize;
    property Ptr[A: TPegtopVirtualAddr]: Pointer read GetPtr;
  end;

  TPegtopCompilerCode = class(TPegtopCompilerData)
  private
    FData: TPegtopCompilerData;
    FNameSpaces: TObjectList;
    FIndent: Integer;
    FLines: TStrings;
    procedure ClearLines;
  protected
    function AddrToHex(A: TPegtopVirtualAddr): String;
  public
    constructor Create(AData: TPegtopCompilerData);
    destructor Destroy; override;
    procedure CleanUp;
    procedure Clear; override;
    procedure AddCode(V: Byte);
    procedure AddAddr(V: LongWord);
    procedure AddLine(S: String);
    procedure IncIndent;
    procedure DecIndent;
    procedure OpenNameSpace(Name: String);
    procedure CloseNameSpace;
    procedure AddLabel(Lab: String);
    procedure AddJump(Lab: String);
    property Lines: TStrings read FLines;
    // property Code: TPegtopCodeSegment read FCode;
    property Data: TPegtopCompilerData read FData;
  end;

implementation

type
  PDouble   = ^Double;
  PInt64    = ^Int64;
  PLongint  = ^Longint;
  PSmallint = ^Smallint;

  TPegtopJump = class
  private
    FLab: String;
    FAdr: TPegtopVirtualAddr;
  public
    constructor Create(ALab: String; AAdr: TPegtopVirtualAddr);
    property Lab: String read FLab;
    property Adr: TPegtopVirtualAddr read FAdr;
  end;

  TPegtopJumpList = class(TObjectList)
  public
    function Find(Lab: String): TPegtopJump;
  end;

  TPegtopNameSpace = class
  private
    FLabels: TPegtopJumpList;
    FJumps: TPegtopJumpList;
    FCode: TPegtopCompilerCode;
    FName: String;
  public
    constructor Create(AName: String; ACode: TPegtopCompilerCode);
    destructor Destroy; override;
    function LabelMissing: Boolean;
    function ListJumps: String;
    procedure AddLabel(Lab: String; Adr: TPegtopVirtualAddr);
    procedure AddJump(Lab: String);
    procedure ResolveJumps(ToLabel: TPegtopJump);
    property Name: String read FName;
  end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopCompilerData
////////////////////////////////////////////////////////////////////////////////

constructor TPegtopCompilerData.Create;
begin
  SetLength(FData, 32);
  FSize := 0;
  FAlign[cdtByte] := 1;
  FAlign[cdtDouble] := 8;
  FAlign[cdtInt64] := 8;
  FAlign[cdtLongint] := 4;
  FAlign[cdtSmallint] := 2;
  FAlign[cdtWord] := 2;
end;

destructor TPegtopCompilerData.Destroy;
begin
  inherited;
end;

procedure TPegtopCompilerData.Clear;
begin
  FSize := 0;
end;

function TPegtopCompilerData.AddByte(V: Byte): TPegtopVirtualAddr;
var
  A: TPegtopVirtualAddr;
begin
  A := IncSize(SizeOf(Byte), FAlign[cdtByte]);
  FData[A] := V;
  Result := A;
end;

function TPegtopCompilerData.AddDouble(V: Double): TPegtopVirtualAddr;
var
  A: TPegtopVirtualAddr;
  P: PDouble;
begin
  A := IncSize(SizeOf(Double), FAlign[cdtDouble]);
  P := @FData[A];
  P^ := V;
  Result := A;
end;

function TPegtopCompilerData.AddInt64(V: Int64): TPegtopVirtualAddr;
var
  A: TPegtopVirtualAddr;
  P: PInt64;
begin
  A := IncSize(SizeOf(Int64), FAlign[cdtInt64]);
  P := @FData[A];
  P^ := V;
  Result := A;
end;

function TPegtopCompilerData.AddLongint(V: Longint): TPegtopVirtualAddr;
var
  A: TPegtopVirtualAddr;
  P: PLongint;
begin
  A := IncSize(SizeOf(LongInt), FAlign[cdtLongInt]);
  P := @FData[A];
  P^ := V;
  Result := A;
end;

function TPegtopCompilerData.AddSmallint(V: Smallint): TPegtopVirtualAddr;
var
  A: TPegtopVirtualAddr;
  P: PSmallint;
begin
  A := IncSize(SizeOf(Smallint), FAlign[cdtSmallint]);
  P := @FData[A];
  P^ := V;
  Result := A;
end;

function TPegtopCompilerData.AddWord(V: Word): TPegtopVirtualAddr;
var
  A: TPegtopVirtualAddr;
  P: PSmallint;
begin
  A := IncSize(SizeOf(Word), FAlign[cdtWord]);
  P := @FData[A];
  P^ := V;
  Result := A;
end;

function TPegtopCompilerData.IncSize(Bytes, Align: Integer): TPegtopVirtualAddr;
begin
  if (FSize mod Align) <> 0 then begin
    FSize := FSize + Align - (FSize mod Align);
  end;
  Result := FSize;
  FSize := FSize + Bytes;
  if FSize > Length(FData) then SetLength(FData, FSize + 32);
end;

procedure TPegtopCompilerData.SetByte(A: TPegtopVirtualAddr; V: Byte);
begin
  FData[A] := V;
end;

function TPegtopCompilerData.GetPtr(A: TPegtopVirtualAddr): Pointer;
begin
  Result := @FData[A];
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopCompilerCode
////////////////////////////////////////////////////////////////////////////////

constructor TPegtopCompilerCode.Create(AData: TPegtopCompilerData);
var
  T: TPegtopCompilerDataType;
begin
  inherited Create;
  FNameSpaces := TObjectList.Create;
  FData := AData;
  FLines := TStringList.Create;
  FIndent := 0;
  for T := Low(TPegtopCompilerDataType) to High(TPegtopCompilerDataType) do
    FAlign[T] := 1; // important!!! (there are no gaps in code)
end;

destructor TPegtopCompilerCode.Destroy;
begin
  FLines.Free; // GRU 1.01
  FNameSpaces.Free;
end;

procedure TPegtopCompilerCode.CleanUp;
begin
  if FNameSpaces.Count <> 0 then raise EPegtopCompiler.Create('All namespaces have to be closed when compilation is done.');
end;

function TPegtopCompilerCode.AddrToHex(A: TPegtopVirtualAddr): String;
begin
  if A < 0 then begin
    if A < -128 then Result := '-' + IntToHex(-A, 8)
    else Result := '-' + IntToHex(-A, 2);
  end
  else begin
    if A >= 128 then Result := '+' + IntToHex(A, 8)
    else Result := '+' + IntToHex(A, 2);
  end;
end;

procedure TPegtopCompilerCode.AddCode(V: Byte);
begin
  AddByte(V);
end;

procedure TPegtopCompilerCode.AddAddr(V: LongWord);
begin
  AddByte(V and 255);
  AddByte(V shr 8 and 255);
  AddByte(V shr 16 and 255);
  AddByte(V shr 24 and 255);
end;

procedure TPegtopCompilerCode.OpenNameSpace(Name: String);
begin
  AddLine(Name + ' begin');
  IncIndent;
  FNameSpaces.Add(TPegtopNameSpace.Create(Name, Self));
end;

procedure TPegtopCompilerCode.CloseNameSpace;
var
  LabelMising: Boolean;
  JumpList: String;
  CurrentNameSpace: TPegtopNameSpace;
begin
  if FNameSpaces.Count > 0 then begin
    CurrentNameSpace := TPegtopNameSpace(FNameSpaces[FNameSpaces.Count-1]);
    LabelMising := CurrentNameSpace.LabelMissing;
    JumpList := CurrentNameSpace.ListJumps;
    DecIndent;
    AddLine(CurrentNameSpace.Name + ' end');
    FNameSpaces.Delete(FNameSpaces.Count-1);
    if LabelMising then raise EPegtopCompiler.CreateFmt('Following labels are missing in this namespace: %s', [JumpList]);
  end
  else begin
    raise EPegtopCompiler.Create('No namespace open.');
  end;
end;

procedure TPegtopCompilerCode.AddLabel(Lab: String);
var
  CurrentNameSpace: TPegtopNameSpace;
begin
  if FNameSpaces.Count > 0 then begin
    DecIndent;
    AddLine(Lab + ':');
    IncIndent;
    CurrentNameSpace := TPegtopNameSpace(FNameSpaces[FNameSpaces.Count-1]);
    CurrentNameSpace.AddLabel(Lab, Size);
  end
  else begin
    raise EPegtopCompiler.Create('No namespace open.');
  end;
end;

procedure TPegtopCompilerCode.AddJump(Lab: String);
var
  CurrentNameSpace: TPegtopNameSpace;
begin
  if FNameSpaces.Count > 0 then begin
    CurrentNameSpace := TPegtopNameSpace(FNameSpaces[FNameSpaces.Count-1]);
    CurrentNameSpace.AddJump(Lab);
  end
  else begin
    raise EPegtopCompiler.Create('No namespace open.');
  end;
end;

procedure TPegtopCompilerCode.AddLine(S: String);
begin
  FLines.Add(StringOfChar(' ', FIndent * 2) + S);
end;

procedure TPegtopCompilerCode.Clear;
begin
  inherited;
  ClearLines;
end;

procedure TPegtopCompilerCode.ClearLines;
begin
  FLines.Clear;
  FIndent := 0;
end;

procedure TPegtopCompilerCode.IncIndent;
begin
  Inc(FIndent);
end;

procedure TPegtopCompilerCode.DecIndent;
begin
  Dec(FIndent);
  if FIndent < 0 then FIndent := 0;
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopJump
////////////////////////////////////////////////////////////////////////////////

constructor TPegtopJump.Create(ALab: String; AAdr: TPegtopVirtualAddr);
begin
  FLab := ALab;
  FAdr := AAdr;
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopJumpList
////////////////////////////////////////////////////////////////////////////////

function TPegtopJumpList.Find(Lab: String): TPegtopJump;
var
  Jump: TPegtopJump;
  I: Integer;
  S: String;
begin
  Result := NIL;
  S := AnsiUpperCase(Lab);
  I := 0;
  while (Result = NIL) and (I < Count) do begin
    Jump := TPegtopJump(Items[I]);
    if AnsiUpperCase(Jump.Lab) = S then Result := Jump
    else Inc(I);
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopNameSpace
////////////////////////////////////////////////////////////////////////////////

constructor TPegtopNameSpace.Create(AName: String; ACode: TPegtopCompilerCode);
begin
  FName := AName;
  FCode := ACode;
  FLabels := TPegtopJumpList.Create;
  FJumps := TPegtopJumpList.Create;
end;

destructor TPegtopNameSpace.Destroy;
begin
  FLabels.Free;
  FJumps.Free;
  inherited;
end;

function TPegtopNameSpace.LabelMissing: Boolean;
begin
  Result := FJumps.Count > 0;
end;

function TPegtopNameSpace.ListJumps: String;
var
  I: Integer;
  Jump: TPegtopJump;
begin
  if FJumps.Count > 0 then begin
    Jump := TPegtopJump(FJumps[0]);
    Result := Jump.Lab;
    for I := 1 to FJumps.Count-1 do begin
      Jump := TPegtopJump(FJumps[I]);
      Result := Result + ', ' + Jump.Lab;
    end;
  end
  else begin
    Result := '';
  end;
end;

procedure TPegtopNameSpace.AddLabel(Lab: String; Adr: TPegtopVirtualAddr);
var
  NewLabel: TPegtopJump;
begin
  if FLabels.Find(Lab) <> NIL then raise EPegtopCompiler.CreateFmt('Label %s does already exist in this namesapce.', [Lab]);
  NewLabel := TPegtopJump.Create(Lab, Adr);
  FLabels.Add(NewLabel);
  ResolveJumps(NewLabel);
end;

procedure TPegtopNameSpace.AddJump(Lab: String);
var
  ToLabel: TPegtopJump;
  NewJump: TPegtopJump;
  Dist: Integer;
begin
  ToLabel := FLabels.Find(Lab);
  if ToLabel = NIL then begin // label does not exist yet
    NewJump := TPegtopJump.Create(Lab, FCode.Size);
    FJumps.Add(NewJump);
    FCode.AddByte(0); // resolve later (see ResolveJumps)
  end
  else begin
    Dist := ToLabel.Adr - FCode.Size;
    if (Dist < -128) or (Dist > 127) then raise EPegtopCompiler.CreateFmt('Label %s out of scope.', [ToLabel.Lab]);
    FCode.AddByte(Dist and $FF);
  end;
end;

procedure TPegtopNameSpace.ResolveJumps(ToLabel: TPegtopJump);
var
  I: Integer;
  Jump: TPegtopJump;
  S: String;
  Dist: Integer;
begin
  S := AnsiLowerCase(ToLabel.Lab);
  for I := FJumps.Count-1 downto 0 do begin
    Jump := TPegtopJump(FJumps[I]);
    if AnsiLowerCase(Jump.Lab) = S then begin
      Dist := ToLabel.Adr - (Jump.Adr + 1);
      if (Dist < -128) or (Dist > 127) then raise EPegtopCompiler.CreateFmt('Label %s out of scope.', [ToLabel.Lab]);
      FCode.SetByte(Jump.Adr, Dist and $FF);
      FJumps.Delete(I);
    end;
  end;
end;

end.
